#ifdef _WIN32
#  define WINDOWS_LEAN_AND_MEAN
#  define NOMINMAX
#  include <windows.h>
#endif

// includes, system
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

// includes, GL
#include <GL/glew.h>

#if defined (__APPLE__) || defined(MACOSX)
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

// includes
#include <cutil.h>
#include <cutil_gl_error.h>
#include <cuda_gl_interop.h>

//#include "screengrab.h"

int width = 256;
int height = 256;
int texSize;

GLuint bufferObj;  // pixel buffer object to use for texturing
GLuint texObj;
unsigned int  *GPU_PBOPtr = NULL;
unsigned char *Img = NULL;

#define BUFFER_OFFSET(i) ((char *)NULL + (i))	
GLuint ioBuf[10];

void OnReshape(int w, int h)
{
	// prevent a division by zero when minimising the window
	if (h==0)
		h=1;

	// set the drawable region of the window
	glViewport(0,0,w,h);

	// set up the projection matrix 
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	// just use a perspective projection
	gluPerspective(45,(float)w/h,0.1,100);

	// go back to modelview matrix so we can move the objects about
	glMatrixMode(GL_MODELVIEW);
}

void OnDraw() 
{

	// clear the screen & depth buffer
	glClear(GL_DEPTH_BUFFER_BIT|GL_COLOR_BUFFER_BIT);

	// clear the previous transform
	glLoadIdentity();

	// set the camera position
	gluLookAt(	0,0,5,	//	eye pos
				0,0,0,	//	aim point
				0,1,0);	//	up direction

    CUDA_SAFE_CALL( cudaGLUnmapBufferObject(bufferObj) );
   
	glBegin( GL_QUADS );
    glTexCoord2d(0.0,0.0); glVertex2d(-1.0,-1.0);
    glTexCoord2d(1.0,0.0); glVertex2d(+1.0,-1.0);
    glTexCoord2d(1.0,1.0); glVertex2d(+1.0,+1.0);
    glTexCoord2d(0.0,1.0); glVertex2d(-1.0,+1.0);
    glEnd();

	// currently we've been drawing to the back buffer, we need
	// to swap the back buffer with the front one to make the image visible
	glutSwapBuffers();
}

void OnInit() 
{
	// enable depth testing
	glEnable(GL_DEPTH_TEST);
	glClearColor(0.4,0.5,0.8,1.0);

	//data reading
	BYTE * data;
	FILE * file;
	// open texture data
	file = fopen( "texture.raw", "rb" );
	if ( file == NULL ) 
	{
		printf("file reading error\n");
		printf("press any key to exit");
		getchar();
		exit(0);
	}
	data = (BYTE*)malloc( width * height * 3 );
	// read texture data
	fread( data, width * height * 3, 1, file );
	fclose( file );

	glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &texObj);
    glBindTexture(GL_TEXTURE_2D, texObj);

    glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,  GL_REPLACE );
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width,height, 0, GL_RGB, GL_UNSIGNED_BYTE,data);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

	free( data );

	glewInit();
	glGenBuffers( 1, &bufferObj );
	//glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, bufferObj);
	//glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB,512*512*4, NULL, GL_STREAM_COPY);
	//cudaGLRegisterBufferObject(bufferObj);
	//CUT_CHECK_ERROR_GL();
 //   printf("PBO created.\n");
}

void OnExit() 
{
	//glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0); 
}

void SaveScreenGrab2(const char* filename) 
{
	// get some info about the screen size
	unsigned sw      =  glutGet(GLUT_WINDOW_WIDTH);
	unsigned sh      =  glutGet(GLUT_WINDOW_HEIGHT);
	unsigned bpp     =  glutGet(GLUT_WINDOW_RGBA) ? 4 : 3;
	GLenum   format  =  (bpp==4) ? GL_RGBA : GL_RGB;

	// allocate memory to store image data
	unsigned char* pdata = new unsigned char[sw*sh*bpp];
	//unsigned char * pdata;

	unsigned int timer = 0;
    float elapsedTimeInMs = 0.0f;
    CUT_SAFE_CALL( cutCreateTimer( &timer ) );
    
	CUT_SAFE_CALL( cutStartTimer( timer));	

	// read from front buffer
	glReadBuffer(GL_FRONT);

	unsigned char* devPtr;
	//CUDA_SAFE_CALL( cudaMalloc( (void**) &devPtr, sizeof(unsigned char) * sw*sh*bpp));

	// read pixel data
	int repeatTimes = 1;
	for(int i = 0; i<repeatTimes; i++)
	{
	  //glReadPixels(0,0,sw,sh,format,GL_UNSIGNED_BYTE,pdata);

	  //glReadPixels (0, 0, sw, sh, format, GL_UNSIGNED_BYTE, BUFFER_OFFSET(0));
	  //pdata = (unsigned char*)glMapBuffer(GL_PIXEL_PACK_BUFFER_ARB, GL_READ_ONLY);   
	  //glUnmapBuffer(GL_PIXEL_PACK_BUFFER_ARB);

		glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, bufferObj);
		glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB,512*512*4, NULL, GL_STREAM_COPY);
		cudaGLRegisterBufferObject(bufferObj);
		CUT_CHECK_ERROR_GL();
		printf("PBO created.\n");
	  
	  cudaGLMapBufferObject((void**)&devPtr, bufferObj);
	  
	  //dim3  grid(1, 1, 1);
	  //dim3  threads(1, 1, 1);
	  //AddConst<<<grid,threads>>>( devPtr, sw, sh );
	  //CUDA_SAFE_CALL(cudaMemset(devPtr, 0, sw*sh*bpp));
	  CUDA_SAFE_CALL( cudaMemcpy(pdata, devPtr, sw*sh*bpp, cudaMemcpyDeviceToHost));
	  
	  cudaGLUnmapBufferObject(bufferObj);
	  CUDA_SAFE_CALL( cudaGLUnregisterBufferObject(bufferObj) );
	  glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0); 
	}

    //get the the total elapsed time in ms
    CUT_SAFE_CALL( cutStopTimer( timer));
    elapsedTimeInMs = cutGetTimerValue( timer);

	printf("it took %f ms to read back\n",elapsedTimeInMs);
	printf("bandwidth = %f mega byte per second\n\n", (sw * sh * bpp * repeatTimes / elapsedTimeInMs)/1024);

	FILE * file;
	file = fopen( "new5.raw", "wb" );
	if ( file == NULL ) exit(0);
	fwrite( pdata, sw * sh * bpp, 1, file );
	fclose( file );

	//printf("I'm here\n");
	// clean up
	//delete [] pdata;

}

void KeyDown(unsigned char k, int x, int y)
{
	// take a screen shot when s is pressed
	if(k=='s')
		SaveScreenGrab2("screenshot.tga");
}

int main(int argc,char** argv) {

	// initialise glut
	glutInit(&argc,argv);

	// request a depth buffer, RGBA display mode, and we want double buffering
	glutInitDisplayMode(GLUT_DEPTH|GLUT_RGBA|GLUT_DOUBLE);

	// set the initial window size
	int windowSize = 512;
	glutInitWindowSize(windowSize,windowSize);

	// create the window
	glutCreateWindow("A basic glut example");

	// set the function to use to draw our scene
	glutDisplayFunc(OnDraw);

	// set the function to handle changes in screen size
	glutReshapeFunc(OnReshape);

	glutKeyboardUpFunc(KeyDown);
	
	// run our custom initialisation
	OnInit();

	// set the function to be called when we exit
	atexit(OnExit);

	// this function runs a while loop to keep the program running.
	glutMainLoop();
	return 0;
}
